@@ -2,9 +2,8 @@
 
 namespace spec\Prophecy\Doubler\Generator;
 
-use PhpSpec\ObjectBehavior;
 use I\Simply;
-
+use PhpSpec\ObjectBehavior;
 use ReflectionClass;
 use ReflectionMethod;
 use ReflectionParameter;
@@ -13,15 +12,17 @@ class ClassMirrorSpec extends ObjectBehavior
 {
     /**
      * @param ReflectionClass  $class
+     * @param ReflectionClass  $parent
      * @param ReflectionMethod $method1
      * @param ReflectionMethod $method2
      * @param ReflectionMethod $method3
      */
     function it_reflects_a_class_by_mirroring_all_its_public_methods(
-        $class, $method1, $method2, $method3
+        $class, $parent, $method1, $method2, $method3
     )
     {
         $class->getName()->willReturn('Custom\ClassName');
+        $class->getParentClass()->willReturn($parent);
         $class->isInterface()->willReturn(false);
         $class->isFinal()->willReturn(false);
         $class->getMethods(ReflectionMethod::IS_ABSTRACT)->willReturn(array());
@@ -29,9 +30,15 @@ class ClassMirrorSpec extends ObjectBehavior
             $method1, $method2, $method3
         ));
 
+        $parent->getName()->willReturn('Custom\ParentClassName');
+
+        $method1->getDeclaringClass()->willReturn($class);
+        $method2->getDeclaringClass()->willReturn($class);
+        $method3->getDeclaringClass()->willReturn($class);
+
         $method1->getName()->willReturn('getName');
-        $method2->getName()->willReturn('isPublic');
-        $method3->getName()->willReturn('isAbstract');
+        $method2->getName()->willReturn('getSelf');
+        $method3->getName()->willReturn('getParent');
 
         $method1->isFinal()->willReturn(false);
         $method2->isFinal()->willReturn(false);
@@ -53,6 +60,15 @@ class ClassMirrorSpec extends ObjectBehavior
         $method2->getParameters()->willReturn(array());
         $method3->getParameters()->willReturn(array());
 
+        if (version_compare(PHP_VERSION, '7.0', '>=')) {
+            $method1->hasReturnType()->willReturn(true);
+            $method1->getReturnType()->willReturn('string');
+            $method2->hasReturnType()->willReturn(true);
+            $method2->getReturnType()->willReturn('self');
+            $method3->hasReturnType()->willReturn(true);
+            $method3->getReturnType()->willReturn('parent');
+        }
+
         $classNode   = $this->reflect($class, array());
         $classNode->shouldBeAnInstanceOf('Prophecy\Doubler\Generator\Node\ClassNode');
         $classNode->getParentClass()->shouldReturn('Custom\ClassName');
@@ -61,8 +77,14 @@ class ClassMirrorSpec extends ObjectBehavior
         $methodNodes->shouldHaveCount(3);
 
         $classNode->hasMethod('getName')->shouldReturn(true);
-        $classNode->hasMethod('isPublic')->shouldReturn(true);
-        $classNode->hasMethod('isAbstract')->shouldReturn(true);
+        $classNode->hasMethod('getSelf')->shouldReturn(true);
+        $classNode->hasMethod('getParent')->shouldReturn(true);
+
+        if (version_compare(PHP_VERSION, '7.0', '>=')) {
+            $classNode->getMethod('getName')->getReturnType()->shouldReturn('string');
+            $classNode->getMethod('getSelf')->getReturnType()->shouldReturn('\Custom\ClassName');
+            $classNode->getMethod('getParent')->getReturnType()->shouldReturn('\Custom\ParentClassName');
+        }
     }
 
     /**
@@ -86,11 +108,18 @@ class ClassMirrorSpec extends ObjectBehavior
         $method->isStatic()->willReturn(false);
         $method->returnsReference()->willReturn(false);
 
+        if (version_compare(PHP_VERSION, '7.0', '>=')) {
+            $method->hasReturnType()->willReturn(false);
+        }
+
         $parameter->getName()->willReturn('...');
         $parameter->isDefaultValueAvailable()->willReturn(true);
         $parameter->getDefaultValue()->willReturn(null);
         $parameter->isPassedByReference()->willReturn(false);
         $parameter->getClass()->willReturn($class);
+        if (version_compare(PHP_VERSION, '5.6', '>=')) {
+            $parameter->isVariadic()->willReturn(false);
+        }
 
         $classNode = $this->reflect($class, array());
 
@@ -120,6 +149,9 @@ class ClassMirrorSpec extends ObjectBehavior
         $method->getName()->willReturn('innerDetail');
         $method->returnsReference()->willReturn(false);
 
+        if (version_compare(PHP_VERSION, '7.0', '>=')) {
+            $method->hasReturnType()->willReturn(false);
+        }
 
         $classNode   = $this->reflect($class, array());
         $classNode->shouldBeAnInstanceOf('Prophecy\Doubler\Generator\Node\ClassNode');
@@ -149,6 +181,10 @@ class ClassMirrorSpec extends ObjectBehavior
         $method->getName()->willReturn('innerDetail');
         $method->returnsReference()->willReturn(false);
 
+        if (version_compare(PHP_VERSION, '7.0', '>=')) {
+            $method->hasReturnType()->willReturn(false);
+        }
+
         $classNode   = $this->reflect($class, array());
         $classNode->shouldBeAnInstanceOf('Prophecy\Doubler\Generator\Node\ClassNode');
         $classNode->getParentClass()->shouldReturn('Custom\ClassName');
@@ -160,6 +196,92 @@ class ClassMirrorSpec extends ObjectBehavior
         $methodNodes['innerDetail']->isStatic()->shouldReturn(true);
     }
 
+    /**
+     * @param ReflectionClass  $class
+     * @param ReflectionMethod $constructMethod
+     * @param ReflectionMethod $destructMethod
+     * @param ReflectionMethod $sleepMethod
+     * @param ReflectionMethod $wakeupMethod
+     * @param ReflectionMethod $toStringMethod
+     * @param ReflectionMethod $callMethod
+     * @param ReflectionMethod $invokeMethod
+     */
+    function it_reflects_allowed_magic_methods($class, $constructMethod, $destructMethod, $sleepMethod, $wakeupMethod, $toStringMethod, $callMethod, $invokeMethod)
+    {
+        $class->getName()->willReturn('Custom\ClassName');
+        $class->isInterface()->willReturn(false);
+        $class->isFinal()->willReturn(false);
+        $class->getMethods(ReflectionMethod::IS_ABSTRACT)->willReturn(array());
+        $class->getMethods(ReflectionMethod::IS_PUBLIC)->willReturn(array(
+            $constructMethod, $destructMethod, $sleepMethod, $wakeupMethod, $toStringMethod, $callMethod, $invokeMethod
+        ));
+
+        $constructMethod->getName()->willReturn('__construct');
+        $destructMethod->getName()->willReturn('__destruct');
+        $sleepMethod->getName()->willReturn('__sleep');
+        $wakeupMethod->getName()->willReturn('__wakeup');
+        $toStringMethod->getName()->willReturn('__toString');
+        $callMethod->getName()->willReturn('__call');
+        $invokeMethod->getName()->willReturn('__invoke');
+
+        $constructMethod->isFinal()->willReturn(false);
+        $destructMethod->isFinal()->willReturn(false);
+        $sleepMethod->isFinal()->willReturn(false);
+        $wakeupMethod->isFinal()->willReturn(false);
+        $toStringMethod->isFinal()->willReturn(false);
+        $callMethod->isFinal()->willReturn(false);
+        $invokeMethod->isFinal()->willReturn(false);
+
+        $constructMethod->isProtected()->willReturn(false);
+        $destructMethod->isProtected()->willReturn(false);
+        $sleepMethod->isProtected()->willReturn(false);
+        $wakeupMethod->isProtected()->willReturn(false);
+        $toStringMethod->isProtected()->willReturn(false);
+        $callMethod->isProtected()->willReturn(false);
+        $invokeMethod->isProtected()->willReturn(false);
+
+        $constructMethod->isStatic()->willReturn(false);
+        $destructMethod->isStatic()->willReturn(false);
+        $sleepMethod->isStatic()->willReturn(false);
+        $wakeupMethod->isStatic()->willReturn(false);
+        $toStringMethod->isStatic()->willReturn(false);
+        $callMethod->isStatic()->willReturn(false);
+        $invokeMethod->isStatic()->willReturn(false);
+
+        $constructMethod->returnsReference()->willReturn(false);
+        $destructMethod->returnsReference()->willReturn(false);
+        $sleepMethod->returnsReference()->willReturn(false);
+        $wakeupMethod->returnsReference()->willReturn(false);
+        $toStringMethod->returnsReference()->willReturn(false);
+        $callMethod->returnsReference()->willReturn(false);
+        $invokeMethod->returnsReference()->willReturn(false);
+
+        $constructMethod->getParameters()->willReturn(array());
+        $destructMethod->getParameters()->willReturn(array());
+        $sleepMethod->getParameters()->willReturn(array());
+        $wakeupMethod->getParameters()->willReturn(array());
+        $toStringMethod->getParameters()->willReturn(array());
+        $callMethod->getParameters()->willReturn(array());
+        $invokeMethod->getParameters()->willReturn(array());
+
+        if (version_compare(PHP_VERSION, '7.0', '>=')) {
+            $constructMethod->hasReturnType()->willReturn(false);
+            $destructMethod->hasReturnType()->willReturn(false);
+            $sleepMethod->hasReturnType()->willReturn(false);
+            $wakeupMethod->hasReturnType()->willReturn(false);
+            $toStringMethod->hasReturnType()->willReturn(false);
+            $callMethod->hasReturnType()->willReturn(false);
+            $invokeMethod->hasReturnType()->willReturn(false);
+        }
+
+        $classNode   = $this->reflect($class, array());
+        $classNode->shouldBeAnInstanceOf('Prophecy\Doubler\Generator\Node\ClassNode');
+        $classNode->getParentClass()->shouldReturn('Custom\ClassName');
+
+        $methodNodes = $classNode->getMethods();
+        $methodNodes->shouldHaveCount(7);
+    }
+
     /**
      * @param ReflectionClass     $class
      * @param ReflectionMethod    $method
@@ -167,9 +289,10 @@ class ClassMirrorSpec extends ObjectBehavior
      * @param ReflectionParameter $param2
      * @param ReflectionClass     $typeHint
      * @param ReflectionParameter $param3
+     * @param ReflectionParameter $param4
      */
     function it_properly_reads_methods_arguments_with_types(
-        $class, $method, $param1, $param2, $typeHint, $param3
+        $class, $method, $param1, $param2, $typeHint, $param3, $param4
     )
     {
         $class->getName()->willReturn('Custom\ClassName');
@@ -183,7 +306,11 @@ class ClassMirrorSpec extends ObjectBehavior
         $method->isProtected()->willReturn(true);
         $method->isStatic()->willReturn(false);
         $method->returnsReference()->willReturn(false);
-        $method->getParameters()->willReturn(array($param1, $param2, $param3));
+        $method->getParameters()->willReturn(array($param1, $param2, $param3, $param4));
+
+        if (version_compare(PHP_VERSION, '7.0', '>=')) {
+            $method->hasReturnType()->willReturn(false);
+        }
 
         $param1->getName()->willReturn('arg_1');
         $param1->isArray()->willReturn(true);
@@ -213,6 +340,22 @@ class ClassMirrorSpec extends ObjectBehavior
         $param3->isPassedByReference()->willReturn(false);
         $param3->allowsNull()->willReturn(true);
 
+        $param4->getName()->willReturn('arg_4');
+        $param4->isArray()->willReturn(false);
+        $param4->getClass()->willReturn($typeHint);
+        $param4->isPassedByReference()->willReturn(false);
+        $param4->allowsNull()->willReturn(true);
+
+        if (version_compare(PHP_VERSION, '5.6', '>=')) {
+            $param1->isVariadic()->willReturn(false);
+            $param2->isVariadic()->willReturn(false);
+            $param3->isVariadic()->willReturn(false);
+            $param4->isVariadic()->willReturn(true);
+        } else {
+            $param4->isOptional()->willReturn(true);
+            $param4->isDefaultValueAvailable()->willReturn(false);
+        }
+
         $classNode   = $this->reflect($class, array());
         $methodNodes = $classNode->getMethods();
         $argNodes    = $methodNodes['methodWithArgs']->getArguments();
@@ -234,6 +377,15 @@ class ClassMirrorSpec extends ObjectBehavior
         } else {
             $argNodes[2]->isOptional()->shouldReturn(false);
         }
+
+        $argNodes[3]->getName()->shouldReturn('arg_4');
+        $argNodes[3]->getTypeHint()->shouldReturn('ArrayAccess');
+        if (version_compare(PHP_VERSION, '5.6', '>=')) {
+            $argNodes[3]->isVariadic()->shouldReturn(true);
+        } else {
+            $argNodes[3]->isOptional()->shouldReturn(true);
+            $argNodes[3]->getDefault()->shouldReturn(null);
+        }
     }
 
     /**
@@ -258,12 +410,24 @@ class ClassMirrorSpec extends ObjectBehavior
         $method->returnsReference()->willReturn(false);
         $method->getParameters()->willReturn(array($param1));
 
+        if (version_compare(PHP_VERSION, '7.0', '>=')) {
+            $method->hasReturnType()->willReturn(false);
+        }
+
         $param1->getName()->willReturn('arg_1');
         $param1->isArray()->willReturn(false);
         if (version_compare(PHP_VERSION, '5.4', '>=')) {
             $param1->isCallable()->willReturn(false);
         }
         $param1->getClass()->willReturn(null);
+
+        if (version_compare(PHP_VERSION, '7.0', '>=')) {
+            $param1->hasType()->willReturn(false);
+        }
+
+        if (version_compare(PHP_VERSION, '5.6', '>=')) {
+            $param1->isVariadic()->willReturn(false);
+        }
         $param1->isDefaultValueAvailable()->willReturn(false);
         $param1->isOptional()->willReturn(false);
         $param1->isPassedByReference()->willReturn(false);
@@ -284,10 +448,11 @@ class ClassMirrorSpec extends ObjectBehavior
      * @param ReflectionMethod    $method
      * @param ReflectionParameter $param1
      * @param ReflectionParameter $param2
+     * @param ReflectionParameter $param3
      * @param ReflectionClass     $typeHint
      */
     function it_marks_passed_by_reference_args_as_passed_by_reference(
-        $class, $method, $param1, $param2, $typeHint
+        $class, $method, $param1, $param2, $param3, $typeHint
     )
     {
         $class->getName()->willReturn('Custom\ClassName');
@@ -301,7 +466,11 @@ class ClassMirrorSpec extends ObjectBehavior
         $method->isProtected()->willReturn(false);
         $method->isStatic()->willReturn(false);
         $method->returnsReference()->willReturn(false);
-        $method->getParameters()->willReturn(array($param1, $param2));
+        $method->getParameters()->willReturn(array($param1, $param2, $param3));
+
+        if (version_compare(PHP_VERSION, '7.0', '>=')) {
+            $method->hasReturnType()->willReturn(false);
+        }
 
         $param1->getName()->willReturn('arg_1');
         $param1->isArray()->willReturn(false);
@@ -309,9 +478,17 @@ class ClassMirrorSpec extends ObjectBehavior
             $param1->isCallable()->willReturn(false);
         }
         $param1->getClass()->willReturn(null);
+        if (version_compare(PHP_VERSION, '5.6', '>=')) {
+            $param1->isVariadic()->willReturn(false);
+        }
         $param1->isDefaultValueAvailable()->willReturn(false);
         $param1->isOptional()->willReturn(true);
         $param1->isPassedByReference()->willReturn(true);
+
+        if (version_compare(PHP_VERSION, '7.0', '>=')) {
+            $param1->hasType()->willReturn(false);
+        }
+
         $param1->allowsNull()->willReturn(false);
         if (defined('HHVM_VERSION')) {
             $param1->getTypehintText()->willReturn(null);
@@ -320,18 +497,39 @@ class ClassMirrorSpec extends ObjectBehavior
         $param2->getName()->willReturn('arg2');
         $param2->isArray()->willReturn(false);
         $param2->getClass()->willReturn($typeHint);
+        if (version_compare(PHP_VERSION, '5.6', '>=')) {
+            $param2->isVariadic()->willReturn(false);
+        }
         $param2->isDefaultValueAvailable()->willReturn(false);
         $param2->isOptional()->willReturn(false);
         $param2->isPassedByReference()->willReturn(false);
+
+        if (version_compare(PHP_VERSION, '7.0', '>=')) {
+            $param2->hasType()->willReturn(false);
+        }
+
         $param2->allowsNull()->willReturn(false);
         $typeHint->getName()->willReturn('ArrayAccess');
 
+        $param3->getName()->willReturn('arg2');
+        $param3->isArray()->willReturn(false);
+        $param3->getClass()->willReturn($typeHint);
+        if (version_compare(PHP_VERSION, '5.6', '>=')) {
+            $param3->isVariadic()->willReturn(true);
+        } else {
+            $param3->isOptional()->willReturn(true);
+            $param3->isDefaultValueAvailable()->willReturn(false);
+        }
+        $param3->isPassedByReference()->willReturn(true);
+        $param3->allowsNull()->willReturn(true);
+
         $classNode   = $this->reflect($class, array());
         $methodNodes = $classNode->getMethods();
         $argNodes    = $methodNodes['methodWithArgs']->getArguments();
 
         $argNodes[0]->isPassedByReference()->shouldReturn(true);
         $argNodes[1]->isPassedByReference()->shouldReturn(false);
+        $argNodes[2]->isPassedByReference()->shouldReturn(true);
     }
 
     /**
@@ -366,6 +564,26 @@ class ClassMirrorSpec extends ObjectBehavior
         $classNode->getMethods()->shouldHaveCount(0);
     }
 
+    /**
+     * @param ReflectionClass  $class
+     * @param ReflectionMethod $method
+     */
+    function it_marks_final_methods_as_unextendable($class, $method)
+    {
+        $class->getName()->willReturn('Custom\ClassName');
+        $class->isInterface()->willReturn(false);
+        $class->isFinal()->willReturn(false);
+        $class->getMethods(ReflectionMethod::IS_ABSTRACT)->willReturn(array());
+        $class->getMethods(ReflectionMethod::IS_PUBLIC)->willReturn(array($method));
+
+        $method->isFinal()->willReturn(true);
+        $method->getName()->willReturn('finalImplementation');
+
+        $classNode = $this->reflect($class, array());
+        $classNode->getUnextendableMethods()->shouldHaveCount(1);
+        $classNode->isExtendable('finalImplementation')->shouldReturn(false);
+    }
+
     /**
      * @param ReflectionClass $interface
      */
@@ -418,6 +636,12 @@ class ClassMirrorSpec extends ObjectBehavior
         $method2->getParameters()->willReturn(array());
         $method3->getParameters()->willReturn(array());
 
+        if (version_compare(PHP_VERSION, '7.0', '>=')) {
+            $method1->hasReturnType()->willReturn(false);
+            $method2->hasReturnType()->willReturn(false);
+            $method3->hasReturnType()->willReturn(false);
+        }
+
         $classNode = $this->reflect(null, array($interface1, $interface2));
 
         $classNode->shouldBeAnInstanceOf('Prophecy\Doubler\Generator\Node\ClassNode');
@@ -472,6 +696,12 @@ class ClassMirrorSpec extends ObjectBehavior
         $method2->getParameters()->willReturn(array());
         $method3->getParameters()->willReturn(array());
 
+        if (version_compare(PHP_VERSION, '7.0', '>=')) {
+            $method1->hasReturnType()->willReturn(false);
+            $method2->hasReturnType()->willReturn(false);
+            $method3->hasReturnType()->willReturn(false);
+        }
+
         $classNode = $this->reflect($class, array());
         $methodNodes = $classNode->getMethods();
         $methodNodes->shouldHaveCount(2);
